"""
A CSL reference frame model which has knowledge about the CSL Setup, and the PUNA Hexapod model.

The CSL Reference Frame Model incorporates a Hexapod PUNA model which is represented by the
Reference Frames HEXUSR, HEXOBJ, HEXMEC, and HEXPLT. A number of methods are defined here that
assume these four reference frames exist in the model and behave like a proper hexapod simulator.
Those methods start with the name `hexapod_`, e.g. `hexapod_goto_zero_position()`.

"""

import numpy as np
from egse.coordinates.refmodel import ReferenceFrameModel

HEXUSR = "hexusr"
HEXMEC = "hexmec"
HEXOBJ = "hexobj"
HEXPLT = "hexplt"
HEXOBUSR = "hexobusr"


class CSLReferenceFrameModel(ReferenceFrameModel):
    """
    The CSL reference Frame Model is a specific reference model that adds convenience methods for manipulating the
    Hexapod PUNA which is part of the overall CSL Setup.
    """

    _DEGREES_DEFAULT = ReferenceFrameModel._DEGREES_DEFAULT

    def _create_obusr(self) -> None:
        """Creates the Object User Reference Frame if it does not exist yet."""

        if HEXOBUSR in self:
            return

        hexusr = self.get_frame(HEXUSR)
        hexobj = self.get_frame(HEXOBJ)

        transformation = hexusr.get_active_transformation_to(hexobj)

        self.add_frame(HEXOBUSR, transformation=transformation, reference=HEXUSR)
        self.add_link(HEXOBUSR, HEXOBJ)

    def hexapod_move_absolute(self, translation: np.ndarray, rotation: np.ndarray, degrees: bool = _DEGREES_DEFAULT):
        """Moves/defines the Object Coordinate System expressed in the invariant User Coordinate System.

        The rotation centre coincides with the Object Coordinates System origin and the movements are controlled with
        translation components first (Tx, Ty, tZ) and then the rotation components (Rx, Ry, Rz).

        Args:
            translation (np.ndarray): Translation vector.
            rotation (np.ndarray): Rotation vector.
            degrees (bool): Indicates whether the rotation angles are specified in degrees, rather than radians.
        """

        self.move_absolute_self(HEXOBUSR, translation, rotation, degrees=degrees)

    def hexapod_move_relative_object(
        self, translation: np.ndarray, rotation: np.ndarray, degrees: bool = _DEGREES_DEFAULT
    ):
        """Moves the object relative to its current position and orientation

        The relative movement is expressed in the object coordinate system.

        Args:
            translation (np.ndarray): Translation vector.
            rotation (np.ndarray): Rotation vector.
            degrees (bool): Indicates whether the rotation angles are specified in degrees, rather than radians.
        """

        self.move_relative_self(HEXOBJ, translation, rotation, degrees=degrees)

    def hexapod_move_relative_user(
        self, translation: np.ndarray, rotation: np.ndarray, degrees: bool = _DEGREES_DEFAULT
    ) -> None:
        """Moves the object relative to its current object position and orientation.

        The relative movement is expressed in the (invariant) user coordinate system.

        Args:
            translation (np.ndarray): Translation vector.
            rotation (np.ndarray): Rotation vector.
            degrees (bool): Indicates whether the rotation angles are specified in degrees, rather than radians.
        """

        self.move_relative_other_local(HEXOBJ, HEXUSR, translation, rotation, degrees=degrees)

    def hexapod_configure_coordinates(
        self,
        usr_trans: np.ndarray,
        usr_rot: np.ndarray,
        obj_trans: np.ndarray,
        obj_rot: np.ndarray,
    ) -> None:
        """Changes the definition of the User Coordinate System and the Object Coordinate System in the hexapod.

        Args:
            usr_trans (np.ndarray): Translation vector used to define the User Coordinate System relative to the Machine Coordinate System.
            usr_rot (np.ndarray): Rotation vector used to define the User Coordinate System relative to the Machine Coordinate System.
            obj_trans (np.ndarray): Translation vector used to define the Object Coordinate System relative to the Platform Coordinate System.
            obj_rot (np.ndarray): Rotation vector used to define the Object Coordinate System relative to the Platoform Coordinate System.
        """

        self.remove_link(HEXUSR, HEXMEC)
        self.remove_link(HEXOBJ, HEXPLT)
        self.get_frame(HEXUSR).set_translation_rotation(usr_trans, usr_rot)
        self.get_frame(HEXOBJ).set_translation_rotation(obj_trans, obj_rot)
        self.add_link(HEXUSR, HEXMEC)
        self.add_link(HEXOBJ, HEXPLT)

    def hexapod_goto_zero_position(self) -> None:
        """Instructs the hexapod to go to its zero position"""

        self.move_absolute_self(HEXPLT, translation=np.array([0, 0, 0]), rotation=np.array([0, 0, 0]))

    def hexapod_goto_retracted_position(self) -> None:
        """Instructs the hexapod to go to its retracted position."""

        self.move_absolute_self(HEXPLT, translation=np.array([0, 0, -20]), rotation=np.array([0, 0, 0]))



import sys
import pandas

from egse.setup import Setup


def laser_tracker_to_dict(file_xls: str, setup: Setup):
    """Reads the laser tracker file and returns a dictionary of reference frames.

    Args:
        file_xls (str): Path to the laser tracker file.  This is an excell sheet provided by CSL.
        setup (Setup): Setup object containing the default reference frames.

    Known Features:
        - no link can be included:
                the links are not in the xls file
                hardcoding them would cause trouble when ingesting a partial model
        - the ReferenceFrames references are included, but they are based
          on a hardcoded model.
          In particular, it is assumed that gliso is the master!

        - a "Master" ReferenceFrame is enforced, with the name "Master" (capital)

        - the names of the reference frames are returned lowercase, without '_'
          ("Master" is an exception)

    Returns:
        Dictionary of reference frames compatible with egse.coordinates.dict_to_ref_model.
    """

    # Predefined model -- gliso ~ master

    """
    predef_refs={}
    predef_refs['gltab']  = 'glfix'
    predef_refs['glfix']  = 'glrot'
    predef_refs['glrot']  = 'gliso'

    predef_refs['gliso']  = 'Master'
    predef_refs['Master'] = 'Master'

    predef_refs['hexiso'] = 'gliso'
    predef_refs['hexmec'] = 'hexiso'
    predef_refs['hexplt'] ='hexmec'
    predef_refs['hexobj'] = 'hexplt'
    predef_refs['hexobusr'] = 'hexusr'
    predef_refs['hexusr'] = 'hexmec'
    predef_refs['fpaaln'] = 'gliso'
    predef_refs['fpasen'] = 'fpaaln'
    predef_refs['fpamec'] = 'fpaaln'
    predef_refs['toumec'] = 'gliso'
    predef_refs['toul6'] = 'toumec'
    predef_refs['toualn'] = 'toumec'
    predef_refs['touopt'] = 'toualn'
    predef_refs['marialn'] = 'toualn'
    predef_refs['cammec'] = 'toualn'
    predef_refs['cambor'] = 'toualn'
    """

    predef_refs = setup.csl_model.default_refs

    # Read input file

    pan = pandas.read_excel(file_xls, sheet_name="Data", usecols="A:D", names=["desc", "x", "y", "z"])

    num_rows = pan.shape[0]

    desc = pan["desc"].values
    colx = pan["x"].values
    coly = pan["y"].values
    colz = pan["z"].values

    reference_frames = dict()
    reference_frames["Master"] = (
        "ReferenceFrame//([0.0000,0.0000,0.0000 | [0.0000,0.0000,0.0000 | Master | Master | [])"
    )

    links = "[]"

    i, frame = -1, -1
    while i < num_rows:
        i += 1

        try:
            frame = desc[i].find("Frame")
        except:
            frame = -1
            continue

        if frame >= 0:
            try:
                name = desc[i][desc[i].find("::") + 2 :].lower().replace("_", "")

                if (desc[i + 2].lower().find("translation") < 0) or (desc[i + 3].lower().find("rotation") < 0):
                    raise Exception(f"Unexpected File Structure after row {i} : {desc[i]}")

                translation = f"[{float(colx[i + 2]):.6f},{float(coly[i + 2]):.6f},{float(colz[i + 2]):.6f}"
                rotation = f"[{float(colx[i + 3]):.6f},{float(coly[i + 3]):.6f},{float(colz[i + 3]):.6f}"

                if name in predef_refs.keys():
                    ref = predef_refs[name]
                else:
                    ref = "None"

                reference_frames[name] = f"ReferenceFrame//({translation} | {rotation} | {name} | {ref} | {links})"

            except:
                print(f"Frame extraction issue after row {i} : {desc[i]}")
                print(sys.exc_info())

    return reference_frames
